#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#define MAGIC_SIZE 8

typedef struct _normal_gpio_cfg
{
    char      port;                       //ex.PA,PB...
    char      port_num;                   //ex.PA0,PA1
    char      mul_sel;                    //function num
    char      pull;                       //
    char      drv_level;                  //
    char      data;                       //
    char      reserved[2];                //
}
normal_gpio_cfg;

/******************************************************************************/
/*               the control information stored in file head                  */
/******************************************************************************/
struct spare_boot_ctrl_head
{
	unsigned int  jump_instruction;   // one intruction jumping to real code
	unsigned char magic[MAGIC_SIZE];  // ="u-boot"
	unsigned int  check_sum;          // generated by PC
	unsigned int  align_size;         // align size in byte
	unsigned int  length;             // the size of all file
	unsigned int  uboot_length;       // the size of uboot
	unsigned char version[8];         // uboot version
	unsigned char platform[8];        // platform information
	int           reserved[1];        //stamp space, 16bytes align
};

/******************************************************************************/
/*                          the data stored in file head                      */
/******************************************************************************/
struct spare_boot_data_head
{
	unsigned int                dram_para[32];
	int                         run_clock;              // Mhz
	int                         run_core_vol;           // mV
	int                         uart_port;              // UART
	normal_gpio_cfg             uart_gpio[2];           // UART GPIO
	int                         twi_port;               // TWI
	normal_gpio_cfg             twi_gpio[2];            // GPIO TWI
	int                         work_mode;
	int                         storage_type;           // 0:nand   1:sdcard  2:spinor
	normal_gpio_cfg             nand_gpio[32];          // nand GPIO
	char                        nand_spare_data[256];   // nand
	normal_gpio_cfg             sdcard_gpio[32];        // sdcard GPIO
	char                        sdcard_spare_data[256]; // sdcard
	int                         secureos_exist;
	int                         dtb_offset;
	int                         reserved[4];            //256bytes align
};

struct spare_boot_ext_head
{
	int data[4];
};


struct spare_boot_head_t
{
	struct spare_boot_ctrl_head    boot_head;
	struct spare_boot_data_head    boot_data;
	struct spare_boot_ext_head     boot_ext[16];
};

void *buf;
int file_len;

#define STAMP_VALUE                     0x5F0A6C39

/* check sum functon from sun4i boot code */
int gen_check_sum(struct spare_boot_ctrl_head *head_p)
{
	uint32_t length;
	uint32_t *buf;
	uint32_t loop;
	uint32_t i;
	uint32_t sum;

	length = head_p->length;
	if ((length & 0x3) != 0)	/* must 4-byte-aligned */
		return -1;
	buf = (uint32_t *)head_p;
	head_p->check_sum = STAMP_VALUE;	/* fill stamp */
	loop = length >> 2;

	/* calculate the sum */
	for (i = 0, sum = 0; i < loop; i++)
		sum += buf[i];

	/* write back check sum */
	head_p->check_sum = sum;

	return 0;
}

static int round_sector(int len)
{
	return (len + 511) & ~511;
}

static void append_image(char *filename, int id)
{
	struct spare_boot_head_t *header;
	int fd;
	int len;
	int copy_offs = file_len;

	fd = open(filename, O_RDONLY);
	len = lseek(fd, 0, SEEK_END);
	lseek(fd, 0, SEEK_SET);
	/* Round up to sector boundary */

	file_len += round_sector(len);
	buf = realloc(buf, file_len);
	memset(buf + copy_offs, 0, round_sector(len));
	assert(read(fd, buf + copy_offs, len) == len);

	*(int *)(buf + (16 * id + 0x500)) = copy_offs;
	*(int *)(buf + (16 * id + 0x504)) = round_sector(len);
}

int main(int argc, char **argv)
{
	struct spare_boot_head_t *header;
	int fd;

	if (argc != 4) {
		printf("Syntax Error. Call me like this:\n"
		       "  %s: <scp> <bl3 with uboot> <output>\n", argv[0]);
		exit(1);
	}

	/* Assemble header */
	file_len = sizeof(struct spare_boot_head_t);
	header = buf = malloc(file_len);

	memset(buf, 0, file_len);
	strcpy(header->boot_head.magic, "uboot");

	/* Append sec firmware */
	append_image(argv[1], 2);

	/* Append ATF firmware which includes U-Boot */
	append_image(argv[2], 0);

	/* Pad image to 0x2000 boundary */
	file_len = (file_len + 0x2000) & ~0x1fff;
	buf = realloc(buf, file_len);

	/* Calculate checksum */
	header = buf;
	header->boot_head.length = file_len;
	gen_check_sum(&header->boot_head);

	fd = open(argv[3], O_CREAT | O_WRONLY | O_TRUNC, S_IRWXU | S_IRGRP | S_IROTH);
	assert(write(fd, buf, file_len) == file_len);
	close(fd);

	return 0;
}
